/************************************************************************
 *   cp - copy files                                                    *
 *                                                                      *
 *   Syntax:     ex cp;'sourcefile targetfile'                          *
 *               ex cp;'sourcefiles targetdirectory'                    *
 *                                                                      *
 *      -v      verbose mode - report on files removed/failed           *
 *      -i      interactively ask about overwriting existing files      *
 *      -k      keep uncomplete files if failure during copy            *
 *      -m      move files rather than simply copying them              *
 *      -r      recursively copy files.                                 *
 *      -s      sort of files.                                          *
 *      -z      only create directories if NOT empty in -r mode         *
 ************************************************************************/

/*========================================================================
                            AMENDMENT HISTORY
                            ~~~~~~~~~~~~~~~~~
25/01/91  v1.00 DJW   - Original version by David J. Walker
01/04/91  v1.01 DJW   - The -i, -k and -v flags added
26/05/91  v1.02 DJW   - Start/stop improved
12/09/92  v1.05 DJW   - Added option to copy multiple files to a target
                        directory.
                      - Removed need for any routines from stdio package.
20/10/92  v1.06 DJW   - Fixed problem in WCOPY mode with getting filenames
                        wrong
30/08/93  v2.00 DJW   - Rebuilt with new version of c68 libraries to be 
                        QXL compatible
                      - Added -m and -r options
                      - Removed command line expansion as this is now done
                        internally within the program as required.
22/10/93  v2.01 DJW   - Corrected problem when target was a directory.
                      - Added -l option (for Phil Borman)
25/12/93  v2.02 DJW   - Changed so that all directories required are created
                        before any files copied when doing recursive copy.
06/04/94  v2.03 DJW   - Fixed problem with incorrect error message about
                        copying a file to itself when on different devices
                      - Fixed problem on level 2+ drivers with setting 
                        dates and version numbers correctly.
-------------------------------------------------------------------------*/

char    _prog_name[] = "cp";
char    _version[]   = " v2.03 ";
char    _copyright[] = " (c) 1991-1994 D. J. Walker ";

#include <qdos.h>
#include <sms.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef QDOS
#include <debug.h>
#else
#define DBG(params)
#define DBG_INIT()
#endif

#ifdef QDOS
struct  WINDOWDEF _condetails = {2, 1, 0, 7, 468, 205, 24, 24}; 
void    (*_consetup)()    = consetup_title;
long    _stack = 6 * 1024;
long    _stackmargin=1024;
#endif /* QDOS */

#define TRANSFER_UNIT    8192
static char    cpbuf[TRANSFER_UNIT];
static int     iflag = 0;              /* Flag for interactive copy mode */
static int     vflag = 0;              /* Flag for verbose mode */
static int     kflag = 0;              /* Flag for "keep after error" mode */
static int     mflag = 0;              /* Flag for MOVE mode */
static int     rflag = 0;              /* Flag for "recursive copy" mode */
static int     zflag = 0;              /* Flag for "not empty directories" mode */

static int     copycount = 0;          /* Count of files copied OK */
static int     copyfail = 0;           /* Count of files copied with error */
static int     dircount = 0;           /* Count of directories copied OK */
static int     dirfail = 0;            /* Count of directories copied with error */
static int     dircreate = 0;          /* Count of new directories created */

static int     todirectory = 0;

static char    *ListName = NULL;       /* Name of file containing file list (if any) */
static FILE    *fpList = NULL;

static char    *sorttext="";

/*=================================================================== USAGE */
static
void    Usage( int argc, char * argv[], char * text)
/*      ~~~~~
 *  Give details of what are accptable parameters
 *--------------------------------------------------------------------------*/
{
    int     x;

    (void) write( 2, "Command line supplied was:\n", 28U);
    for  (x=0 ; x < argc ; x++) {
        (void) write(2, argv[x], (unsigned) strlen(argv[x]));
        (void) write(2, " ", 1U);
    }

    (void) write (2, "\n\n",2U);
    (void) write (2, text, (unsigned)strlen(text));
    (void) write (2, "\n\n",2U);
    (void) write (2, "Usage:  cp [-[i][k][m][v][s sortorder]] sourcefile targetfile\n",62U);
    (void) write (2, "  or    cp [-[i][k][m][r][v][z][s sortorder]] sourcelist targetdirectory\n",73U);
    (void) write (2, "    -i     query before overwriting\n", 36U);
    (void) write (2, "    -k     keep incomplete files\n", 33U);
    (void) write (2, "    -m     move files\n", 22U);
    (void) write (2, "    -r     recursively copy sub-directories\n", 44U);
    (void) write (2, "    -v     verbose mode\n", 25U);
    (void) write (2, "    -z     do not copy enpty directories\n", 41U);
#if 0
    (void) write (2, "    -l listfile    name of file containing a\n", 46U);
    (void) write (2, "                   a list of filenames to copy\n", 48U);
#endif
    (void) write (2, "    -s sortorder   sort names before copying\n", 45U);
    (void) write (2, "           where sortorder is any combination of:\n", 50U);
    (void) write (2, "             N or n   sort on name\n", 35U);
    (void) write (2, "             U or u   sort on useage\n", 37U);
    (void) write (2, "             S or s   sort on size\n", 35U);
    (void) write (2, "             D or d   sort on date\n", 35U);
    (void) write (2, "             T or t   sort on time\n", 35U);
    (void) write (2, "           Upper case =ascending, lower case=descending\n", 56U);
    exit(-1);
}


/*================================================================= CONFIRM */
static
int     Confirm (char * target)
/*      ~~~~~~~
 *--------------------------------------------------------------------------*/
{
    char t;

    (void) write (2, target, (unsigned)strlen(target));
    (void) write (2, " already exists.  Overwrite (y/n) ",34U);
    (void) read (0, &t, 1U);
    (void) write(2, "\n", 1U);
    return (t == 'y' || t == 'Y');
}


/*======================================================== CHECK_DIRECTORY */
static
int     Check_Directory (char * dirname, int create_flag)
/*      ~~~~~~~~~~~~~~~
 *  This routine is used to check for the existence of a directory.
 *  If the create_flag is set, and the directory does not exist, then
 *  an attempt is made to create it. 
 *--------------------------------------------------------------------------*/
{
    struct stat sbuf;
    char    dirname2[50];
    size_t  sz;
    int     reply;

    DBG(("CHECK_DIRECTORY",0x41,"Enter: Check_Directory(%s, %d)", dirname2, create_flag));
    /*
     *  See if file already exists.
     */
    if (stat(dirname, &sbuf) == 0) {
        DBG(("CHECK_DIRECTORY",0x41,"File %s already exists", dirname));
        DBG(("CHECK_DIRECTORY",0x41,"Exit 1"));
        return 1;
    }
    (void) strcpy (dirname2, dirname);
    sz = strlen (dirname2) - 1;
    if (dirname2[sz] == '_') {
        dirname2[sz] = '\0';
        if (stat(dirname2, &sbuf) == 0) {
            DBG(("CHECK_DIRECTORY",0x41,"File %s already exists", dirname));
            DBG(("CHECK_DIRECTORY",0x41,"Exit 1"));
            return 1;
        }
    }
    if (create_flag != 0) {
        DBG(("CHECK_DIRECTORY",0x41,"creating directory %s", dirname2));
        if ((reply=mkdir (dirname2, 0)) == 0) {
            DBG(("CHECK_DIRECTORY",0x41,"directory %s created OK", dirname2));
            dircreate++;
            if (vflag) {
                (void) write (2, "directory '", 12U);
                (void) write (2, dirname2, strlen(dirname2));
                (void) write (2, "' created\n", 10U);
            }
        }
        DBG(("CHECK_DIRECTORY",0x41,"Exit 1"));
        return 1;
    }
    DBG(("CHECK_DIRECTORY",0x41,"Exit 0"));
    return 0;
}


/*============================================================= COPY_FILE */
static
void    Copy_File (char * source, char * target)
/*      ~~~~~~~~~
 *  This routine handles the copying of single file.  The source
 *  and target names should be set up correctly (including any
 *  path information).
 *---------------------------------------------------------------------*/
{
    register int n, m;
    register int  fd1, fd2;
#ifdef QDOS
    struct qdirect qbuf;
    chanid_t  chid1, chid2;
    long    version,update,backup;
    int     reply;
#endif /* QDOS */

    DBG(("COPY_FILE",0x24,"target filename is %s", target));

    /*
     *  If the verbose flag is set, then we need
     *  to inform the user what is happening.
     */
    if (vflag) {
        if (mflag) {
            (void) write (2, "MOVING ", 7U);
        } else {
            (void) write (2, "COPYING ", 8U);
        }
        (void) write (2, source, (unsigned)strlen(source));
        (void) write (2, "' TO '", 6U);
        (void) write (2, target, (unsigned)strlen(target));
        (void) write (2, "\n", 1U);
    }
    /*
     *  Open source file
     */
    fd1 = open(source, O_RDONLY);
    if ( fd1 < 0) {
        DBG(("COPY_FILE",0x21,"Exit: failed to open source file"));
        (void) write (2, source, (unsigned)strlen(source));
        (void) write (2, " - cannot open file\n", 20U);
        copyfail++;
        return;
    }
    /*
     *   Check if target file exists
     */
    fd2 = open (target, O_RDONLY);
    if (fd2 > 0) {
        DBG(("COPY_FILE",0x28,"... target file already exists"));
        (void) close (fd2);
        if (iflag) {
            if (!Confirm(target)) {
                DBG(("COPY_FILE",0x21,"Exit: user failed to confirm copy"));
                (void) close (fd1);
                copyfail++;
                return;
            }
            DBG(("COPY_FILE",0x28,"... user confirmed copy"));
        }
        (void) unlink (target);
        DBG(("COPY_FILE",0x28,"... so file deleted"));
    }
    /*
     *  Create the output file
     */
    fd2 = open (target, O_WRONLY|O_CREAT|O_TRUNC);
    if (fd2 < 0) {
        (void) close (fd1);
        (void) write (2, target, (unsigned)strlen(target));
        (void) write (2, " - cannot create file\n", 22U);
        DBG(("COPY_FILE",0x21,"Exit: failed to open output file"));
        copyfail++;
        return;
    }
    /*
     *  Copy file contents
     */
    for ( n = 1 ; n ; ) {
        switch ( n = read(fd1, cpbuf, TRANSFER_UNIT)) {
            case -1: /* error */
                DBG(("COPY_FILE",0x21,"Exit: failure reading source file (errno=%d, _oserr=%d)", errno, _oserr));
                (void) write (2, "read error\n", 11U);
                (void) close(fd1);
                (void) close(fd2);
                if (! kflag) {
                    (void) unlink(target);
                    DBG(("COPY_FILE",0x21,"... so deleted source file"));
                }
                copyfail++;
                return;
            case 0: /* EOF */
                break;
            default:
                switch (m = write(fd2, cpbuf, n)) {
                    default:
                        if (m == n) 
                            continue;
                        /* FALLTHRU */
                    case -1:
                        DBG(("COPY_FILE",0x21,"Exit: failure writing target file (errno=%d, _oserr=%d)", errno, _oserr));
                        (void) write (2, "write error\n", 12U);
                        (void) close(fd1);
                        (void) close(fd2);
                        if (! kflag) {
                            (void) unlink(target);
                            DBG(("COPY_FILE",0x21,"... so deleted source file"));
                        }
                        copyfail++;
                        return;
                } /* end of write switch */
        } /* end of read switch */
    } /* end of for loop */
    DBG(("COPY_FILE",0x28," ... copy completed OK"));

    /*
     *  This next part is QDOS specific where we try
     *  and copy the file header information.
     */
    chid1 = getchid(fd1);
    chid2 = getchid(fd2);
    if (fs_headr (chid1, (timeout_t)-1L,(char *)&qbuf,64) != 64) {
        (void) write (2, source, (unsigned)strlen(source));
        (void) write (2, " - failed to read file header\n", 30U);
        (void) close (fd1);
        (void) close (fd2);
        DBG(("COPY_FILE",0x21,"Exit: failed to ready source file header"));
        if (! kflag) {
            (void) unlink(target);
            DBG(("COPY_FILE",0x21,"... so deleted source file"));
        }
        copyfail++;
        return;
    }
    if ((reply=fs_heads (chid2, (timeout_t)-1L,(char *)&qbuf,14)) != 14) {
        (void) write (2, target, (unsigned)strlen(target));
        (void) write (2, " - failed to write file header\n", 31U);
        (void) close (fd1);
        (void) close (fd2);
        DBG(("COPY_FILE",0x21,"Exit: failed to write target file header - reply=%d", reply));
        if (! kflag) {
            (void) unlink(target);
            DBG(("COPY_FILE",0x21,"... so deleted source file"));
        }
        copyfail++;
        return;
    }
    /*
     *  Another QDOS specific part where we try and
     *  (possibly) maintain the date fields the same.
     */
    DBG(("COPY_FILE",0x21, "update=%ld, refdate=%ld, backup=%ld", \
                            qbuf.d_update, qbuf.d_refdate, qbuf.d_backup));
    update = backup = version = -1;
    if ((reply=iof_date(chid1, (timeout_t)-1, 0,&update)) == 0) {
        /*
         *  If we were able to read the update date, use the
         *  same mechanism to set the other fields
         */
        (void) iof_date (chid2, (timeout_t)-1, 0,&update);
        DBG(("COPY_FILE",0x28,"Update date preserved"));
        (void) iof_date (chid1, (timeout_t)-1, 2, &backup);
        (void) iof_date (chid2, (timeout_t)-1, 2, &backup);
        (void) iof_vers (chid1, (timeout_t)-1, &version);
        (void) iof_vers (chid2, (timeout_t)-1, &version);
        (void) close(fd2);
    } else {
        /*
         *  If the call to set the file date failed, then
         *  we try and get it right instead by manipulating
         *  the system clock.
         */
        time_t savetime;
        DBG(("COPY_FILE",0x21,"... all dates not preserved, reply=%d", reply));
        DBG(("COPY_FILE",0x21," trying to set update date by manipulating clock"));
        savetime = mt_rclck();
        mt_sclck(qbuf.d_update);
        (void) close(fd2);
        mt_sclck(savetime);
    }
    (void) close(fd1);
    if ( mflag) {
        DBG(("COPY_FILE",0x28,"removing source file %s",source));
        (void) unlink (source);
    }
    DBG(("COPY_FILE",0x21,"Exit: copy completed"));
    copycount++;
    return;
}


/*========================================================= COPY_CONTROL */
static
void    Copy_Control (char * source, char * target, int create_flag)
/*      ~~~~~~~~~~~~
 *  Control the main copy process.
 *
 *  We have to allow for the fact that the source filename may include
 *  wild card characters.
 *
 *  If the source is a directory then (as long as target is a
 *  directory and -r flag is used we will need to recurse
 *  down a level.
 *-----------------------------------------------------------------------*/
{
    struct DIR_LIST *dirlist, *dptr;
    struct stat sbuf;
    char    SourceDevice[50];
    char    TargetDevice[50];
    char    SourcePath[50];
    char    SourceName[50];
    char    TargetName[50];
    size_t  SourcePathLength;
    size_t  SourceDeviceLength;
    size_t  TargetDeviceLength;
    char    *ptr, *ptr1, *ptr2;
    int     directoryOK = 0;

    DBG(("COPY_CONTROL", 0x11, "Enter: Copy_Control(%s, %s, %d)", \
                                source, target, create_flag));
    if (zflag == 0) {
        if (todirectory > 0 
        &&  (directoryOK=Check_Directory(target, create_flag)) == 0) {
            DBG(("COPY_CONTROL",0x21,"Check_Directory(%s,%d) failed", \
                                            target, create_flag));
            dirfail++;
            return;
        }
    }
    (void) strcpy (SourceDevice, source);
    if ((ptr=strchr(SourceDevice, '_')) == NULL) {
        SourceDevice[0] = '\0';
    } else {
        ptr++;
        *ptr = '\0';
        DBG(("COPY_CONTROL",0x48,"Checking '%s' for device",SourceDevice));
        if (isdirdev(SourceDevice) == 0) {
            SourceDevice[0] = '\0';
        }
    }
    (void) strcpy (TargetDevice, target);
    if ((ptr=strchr(TargetDevice, '_')) == NULL) {
        TargetDevice[0] = '\0';
    } else {
        ptr++;
        *ptr = '\0';
        DBG(("COPY_CONTROL",0x48,"Checking '%s' for device",TargetDevice));
        if (isdirdev(TargetDevice) == 0) {
            TargetDevice[0] = '\0';
        }
    }
    DBG(("COPY_CONTROL",0x24,"SourceDevice=%s, TargetDevice=%s", SourceDevice, TargetDevice));
    SourceDeviceLength = strlen(SourceDevice);
    TargetDeviceLength = strlen(TargetDevice);
    target = &target[TargetDeviceLength];
    DBG(("COPY_CONTROL",0x24,"SourceDeviceLength=%d, TargetDeviceLength=%d", SourceDeviceLength, TargetDeviceLength));

    (void) strcpy (SourcePath, &source[SourceDeviceLength]);
    for (ptr1=SourcePath, ptr2=target ; *ptr1 && *ptr2 ; ptr1++, ptr2++) {
        if ((toupper(*ptr1)) != (toupper(*ptr2)) ) {
            break;
        }
    }
    *ptr1 = '\0';
    SourcePathLength = strlen(SourcePath);
    DBG(("COPY_CONTROL",0x24,"SourcePath=%s, SourcePathLength=%d", SourcePath, SourcePathLength));
    /*
     *  If we are doing a recursive copy, then we
     *  find any directories and create them.
     */
    if (create_flag) {
        if (rflag) {
            DBG(("COPY_CONTROL",0x28,"Checking %s for directories",source));
            dirlist = qdir_read(source,sorttext,4);
            if (dirlist != NULL) {
                DBG(("COPY_CONTROL",0x28,"... directories found"));
                if (todirectory  && directoryOK == 0
                && (directoryOK=Check_Directory (target, create_flag)) == 0) {
                    DBG(("COPY_CONTROL",0x21,"Check_Directory(%s,%d) failed", \
                                                target, create_flag));
                    dirfail++;
                    return;
                }
                for (dptr = dirlist ; dptr ; dptr = dptr->dl_next) {
                    DBG(("COPY_CONTROL",0x28,"Source file %s, type=%d, size=%ld", \
                            dptr->dl_cname, dptr->dl_dir.d_type,dptr->dl_dir.d_length));
                    (void) strcpy (SourceName, dptr->dl_cname);
                    if (SourceName[strlen(SourceName)-1] != '_') {
                        (void) strcat (SourceName, "_");
                    }
                    (void) strcat (SourceName, "*");

                    (void) strcpy (TargetName, TargetDevice);
                    (void) strcat (TargetName, target);
                    if (todirectory) {
                        (void) strcat (TargetName, &dptr->dl_cname[SourceDeviceLength + SourcePathLength]);
                        if (TargetName[strlen(TargetName)-1] != '_') {
                            (void) strcat (TargetName, "_");
                        }
                    }
                    DBG(("COPY_CONTROL",0x48,"calling Copy_Control(%s,%s,%d)", \
                                        SourceName, TargetName, 1));
                    Copy_Control (SourceName, TargetName, 1);
                }
                /*
                 *  Release the memory that was used to
                 *  hold the directory list.
                 */
                qdir_delete (dirlist);
            }
        }
        return;
    }
    /*
     *  See if we have any standard names that
     *  match source mask.
     */
    DBG(("COPY_CONTROL",0x28,"Checking %s for files",source));
    dirlist = qdir_read(source, sorttext, 1+2);
    if (dirlist == NULL) {
        /*
         * It is only OK to not get a match if the
         * target was a directory.
         */
        if (todirectory == 0) {
            (void) write( 2, source, (unsigned)strlen(source));
            (void) write( 2, " not found\n", 11U);
        }
        DBG(("COPY_CONTROL",0x24, "No files found matching mask %s", source));
    } else {
        /*
         *  If we are copying multiple,files, then
         *  ensure that target is a directory.
         */
        DBG(("COPY_CONTROL",0x24,"todirectory=%d", todirectory));
        if (dirlist->dl_next != NULL  && !todirectory) {
            (void) write (2, target, (unsigned)strlen(target));
            (void) write (2, " is not a directory\n", 20U);
            DBG(("MAIN",0x21,"%s is not a directory", target));
            exit (ERR_BN);
        }
        /*
         *  If we are not sure if directory is OK, then
         *  we need to check it.
         */
        if (todirectory > 0 && directoryOK == 0
        &&  (directoryOK=Check_Directory (target, create_flag)) == 0) {
            DBG(("COPY_CONTROL",0x21,"Check_Directory(%s,%d) failed", \
                                        target, create_flag));
            dirfail++;
            return;
        }
        /*
         *  Now copy each file in turn.
         */
        for (dptr = dirlist ; dptr ; dptr = dptr->dl_next) {
            DBG(("COPY_CONTROL",0x21,"Next file = %s, target=%s",dptr->dl_cname,target));
            /*
             *  We have got here, so we are handling a simple
             *  file as the source.
             *
             *  Build up the target filename.
             */
            (void) strcpy (TargetName, TargetDevice);
            (void) strcat (TargetName, target);
            /*
             *  Check that source and target are not the same !
             */
            if (!stricmp(dptr->dl_cname, TargetName)) {
                DBG(("COPY_CONTROL",0x21,"Exit: source and target are the same"));
                (void) write(2, dptr->dl_cname, (unsigned)strlen(dptr->dl_cname));
                (void) write(2, " - cannot copy a file to itself\n", (unsigned) 32U);
                return;
            }
            /*
             *  Check if the source file exists.
             */
            if (stat(dptr->dl_cname, &sbuf)) {
                DBG(("COPY_CONTROL",0x21,"Exit: unable to access source"));
                (void) write (2, dptr->dl_cname, (unsigned)strlen(dptr->dl_cname));
                (void) write (2, " - unable to access source file\n", 31U);
                return;
            }
            if (todirectory) {
                (void) strcat (TargetName, &dptr->dl_cname[SourceDeviceLength + SourcePathLength]);
            }

            DBG(("COPY_CONTROL",0x48,"calling Copy_File(%s,%s)", dptr->dl_cname, TargetName));
            Copy_File(dptr->dl_cname, TargetName);
        } /* End of for loop for files */
        /*
         *  Release the memory that was used to
         *  hold the file list.
         */
        qdir_delete (dirlist);
    }
    /*
     *  If we are doing a recursive copy, then we
     *  find any directories and go down a level.
     */
    if (rflag) {
        DBG(("COPY_CONTROL",0x28,"Checking %s for directories",source));
        dirlist = qdir_read(source,sorttext,4);
        if (dirlist != NULL) {
            DBG(("COPY_CONTROL",0x28,"... directories found"));
            if (todirectory  && directoryOK == 0
            && (directoryOK=Check_Directory (target, create_flag)) == 0) {
                DBG(("COPY_CONTROL",0x21,"Check_Directory(%s,%d) failed", \
                                            target, create_flag));
                dirfail++;
                return;
            }
            for (dptr = dirlist ; dptr ; dptr = dptr->dl_next) {
                DBG(("COPY_CONTROL",0x28,"Source file %s, type=%d, size=%ld", \
                        dptr->dl_cname, dptr->dl_dir.d_type,dptr->dl_dir.d_length));
                (void) strcpy (SourceName, dptr->dl_cname);
                if (SourceName[strlen(SourceName)-1] != '_') {
                    (void) strcat (SourceName, "_");
                }
                (void) strcat (SourceName, "*");

                (void) strcpy (TargetName, TargetDevice);
                (void) strcat (TargetName, target);
                if (todirectory) {
                    (void) strcat (TargetName, &dptr->dl_cname[SourceDeviceLength + SourcePathLength]);
                    if (TargetName[strlen(TargetName)-1] != '_') {
                        (void) strcat (TargetName, "_");
                    }
                }
                DBG(("COPY_CONTROL",0x48,"calling Copy_Control(%s,%s,%d)", \
                                    SourceName, TargetName, 0));
                Copy_Control (SourceName, TargetName, 0);
                dircount++;
            }
            /*
             *  Release the memory that was used to
             *  hold the directory list.
             */
            qdir_delete (dirlist);
        }
    }
    DBG(("COPY_CONTROL",0x11,"Exit:  Copy_Control( %s, %s, %d)", \
                                source, target, create_flag));
    return;
}


/*================================================================= MAIN */
extern
void    main (argc, argv)
/*      ~~~~
 *  This is the starting point for this program.   Basically in this
 *  routine we handle the parameters provided by the user.
 *-----------------------------------------------------------------------*/
int argc;
char *argv[];
{
    extern  int  optind;
    extern  char *optarg;
    char    sourcename[50];
    char    targetname[50];
    char    *ptr;
    int     opt, i, reply;
    struct stat sbuf;

#ifdef TESTING
    DBG (("MAIN", 0x28, "argc=%d", argc));
    for (i = 0 ; i < argc ; i++) {
        DBG(("MAIN",0x28,"argv[%d]=%s", i, argv[i]));
    }
#endif /* TESTING */
    /*
     *  parse any options supplied
     */
    while ((opt=getopt(argc, argv,"iIkKl:L:mMs:S:rRvVzZ")) != EOF ) {
        DBG(("MAIN",0x18,"... opt=%c, optarg=%s, optind=%d",opt,optarg,optind));
        switch (opt) {
        case 'i':       /* interactive mode */
        case 'I':
                iflag++;
                break;
        case 'k':       /* keep incomplete files if copy failes */
        case 'K':
                kflag++;
                break;
        case 'l':
        case 'L':
                ListName = optarg;
                break;
        case 'm':       /* Move (delete after copy succeeeds) */
        case 'M':
                mflag++;
                break;
        case 'r':       /* Recurse down sub-directories */
        case 'R':
                rflag++;
                break;
        case 's':
        case 'S':
                sorttext = optarg;
                DBG(("MAIN",0x11,"Sort text set to %s",sorttext));
                break;
        case 'v':       /* Verbose mode */
        case 'V':
                (void) write (2, _prog_name, (unsigned)strlen(_prog_name));
                (void) write (2, _version, (unsigned)strlen(_version));
                (void) write (2, _copyright, (unsigned)strlen(_copyright));
                (void) write(2, "\n\n", 2U);
                vflag++;
                break;
        case 'z':       /* Do not recreate empty directories */
        case 'Z':
                zflag++;
                break;
        case '-':
                optind++;
                goto END_OPTIONS;
        case '?' :
                Usage (argc, argv, "invalid parameter option");

        } /* end of switch */
    } /* end of while */
END_OPTIONS:
    DBG(("MAIN",0x14,"...finished parsing parameters.  argc=%d, optind=%d",\
                                        argc, optind));

    /*
     *  Some parameter validity checks
     */
    if (zflag != 0 && rflag == 0) {
        Usage (argc, argv, "-z option only allowed if -r option also specified");
    }
    /*
     * now handle the filename arguments
     */

    switch (argc - optind) {
    case -1:
    case 0:
             Usage(argc, argv, "insufficient parameters");
    case 1: /* Wrong number of parameters */
            if (ListName == NULL) {
                 Usage(argc, argv, "insufficient parameters");
            }
            /* FALLTHRU */
    default:
            /*
             *  Check that target exists
             */
            (void) strcpy (targetname, argv[argc - 1]);
            if ((reply=stat(targetname, &sbuf)) == 0) {
                DBG(("MAIN",0x24,"file %s found - st_mode=%d",targetname,sbuf.st_mode));
                todirectory = (sbuf.st_mode & S_IFDIR);
            } else {
                DBG(("MAIN",0x24,"stat() of %s failed - reply=%d, errno=%d, _oserr=%d", \
                                targetname, reply, errno, _oserr));
                /*
                 *  If check fails, and name already ends in an
                 *  underscore, then it is not clear whether we 
                 *  have a (pseduo) directory or not.
                 */
                if ( targetname[strlen(targetname)-1] == '_') {
                    DBG(("MAIN",0x21,"Targetname already ends in underscore"));
                    todirectory = -1;
                } else {
                    /*
                     *  If there is no underscore on the end, then
                     *  add one and try again - matters on some
                     *  systems as dirctory need trailing underscore
                     */
                    (void) strcat (targetname, "_");
                    DBG(("MAIN",0x24,"Added underscore to give %s", targetname));
                    if (stat(targetname, &sbuf) != 0
                    ||  sbuf.st_mode != S_IFDIR) {
                        DBG(("MAIN",0x21,"Failed to access target '%s': errno=%d, _oserr=%d, st_mode=%d", \
                                    targetname, errno, _oserr, sbuf.st_mode));
                        (void) strcpy (targetname, argv[argc - 1]);
                        todirectory = 0;
                    } else {
                        todirectory = 1;
                    }
                }
            }
            DBG(("MAIN",0x24,"todirectory=%d", todirectory));
            /*
             *  Target is a directory.  Even if it does
             *  not end in an underscore, add one for
             *  later convenience.
             */
            if ( todirectory && targetname[strlen(targetname)-1] != '_') {
                (void) strcat (targetname, "_");
                DBG(("MAIN",0x24,"Is a directory, but underscore added to give %s",targetname));
            }
            /*
             *  Now we need to work through each of the user
             *  specified files (on the command line) in turn
             *  copying it to the target
             */
            for (i = optind; i < (argc - 1) ; i++) {
                DBG(("MAIN",0x12,"Copy %s to %s",argv[i],targetname));
                if (todirectory) {
                     Copy_Control(argv[i], targetname, 1);
                }
                 Copy_Control(argv[i], targetname, 0);
            }
            /*
             *  Now we try to see if any filenames were provided
             *  in a file via the -l option.
             */
            if (ListName != NULL) {
                DBG(("MAIN",0x12,"Listfile supplied"));
                if ((fpList = fopen(ListName, "r")) == NULL) {
                    DBG(("MAIN",0x12,"... but open failed"));
                    (void) write (2, "\nUnable to open listfile '", 16U);
                    (void) write (2, ListName, strlen(ListName));
                    (void) write (2, "'\n\n", 3U);
                } else {
                    DBG(("MAIN",0x12,"... opened OK"));
                    while ((ptr=fgets (sourcename, strlen(sourcename), fpList)) != NULL) {
                        DBG(("MAIN",0x28,"reply to fgets() = $%p '%s', errno=%d", ptr, ptr, errno));
                        ptr = &sourcename[strlen(sourcename)-1];
                        if ( *ptr == '\n') {
                            *ptr = '\0';
                        }
                        DBG(("MAIN",0x12,"Copy %s to %s",sourcename,targetname));
                        if (todirectory) {
                            Copy_Control(sourcename, targetname, 1);
                        }
                        Copy_Control(sourcename, targetname, 0);
                    }
                }
            }
            break;
    }   /* end of switch */
    if (vflag) {
        char countbuf[10];
        DBG(("MAIN",0x24,"copycount=%d",copycount));
        (void) itoa(copycount, countbuf);
        if (mflag) {
            (void) write (2, "Files moved OK = ", 17U);
        } else {
            (void) write (2, "Files copied OK = ", 18U);
        }
        (void) write (2, countbuf, (unsigned)strlen(countbuf));
        if (copyfail != 0) {
            DBG(("MAIN",0x24,"copyfail=%d",copyfail));
            (void) itoa(copyfail, countbuf);
            (void) write (2, "\nFiles copies failed = ",23U);
            (void) write (2, countbuf, (unsigned)strlen(countbuf));
        }
        if (rflag) {
            DBG(("MAIN",0x24,"dircount=%d",dircount));
            if (dircreate) {
                (void) itoa(dircreate, countbuf);
                (void) write (2, "\nDirectories created = ", 23U);
                (void) write (2, countbuf, (unsigned)strlen(countbuf));
            }
            (void) itoa(dircount, countbuf);
            (void) write (2, "\nDirectories copied OK = ", 25U);
            (void) write (2, countbuf, (unsigned)strlen(countbuf));
            if (dirfail != 0) {
                DBG(("MAIN",0x24,"dirfail=%d",dirfail));
                (void) itoa(dirfail, countbuf);
                (void) write (2, "\nDirectories failed = ",22U);
                (void) write (2, countbuf, (unsigned)strlen(countbuf));
            }
        }
        (void) write (2, "\n", 1U);
    }
    DBG(("MAIN", 0x21, "Exit:  copyfail=%d, copycount=%d", copyfail, copycount));
    exit (copyfail);
}

