/*

 shexec_c

 Program Execution

 Date		Author		Comments
 ----		------		--------
 08/09/1991	P.J. Taylor	Original.
 23/09/1991	P.J. Taylor	Amended.

 1.04
 ----
 07/03/1998	A.D.Ives	Source code reformatted.
				Error messages shortened.
				retcode changed to global variable gbRetCode
				so that it can be accessed as an internal
				variable.

 1.06
 ----
 11/03/1998	A.D.Ives	global gbRetCode replaced by static in new
				module suppt. Value accessed via Get/Set
				function SI().

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <debug.h>
#include <qdos.h>
#include "sh_h"
#include "suppt_h"
#include "libpjt.h"

#define JOBID_U union jobid_u

JOBID_U
{ jobid_t full ;    /* full job id */
  struct
  { short tag ;     /* tag part */
    short entno ;   /* job table entry number part */
  } part ;
} ;

#define PCH_LAST 1	/* Last in pipe chain		*/
#define PCH_FIRST 2	/* First in pipe chain		*/
#define PCH_FORK 4	/* Last in chain is a fork	*/

void execcmds(CMDTAB *ct, int rep) {

/*

  Execute Commands

  ct		Command Table
  rep		TRUE if result of execution is to be reported to stdout

*/

  int cno,		/* Current command number		*/
      pcno,		/* Command number for command at start of
			   current pipe chain			*/
      pstat,		/* Pipe chain status flag		*/
      i;
  jobid_t owner ,	/* Current owner Job ID 		*/
	  jid , 	/* Current job's Job ID                 */
	  pejob ;	/* ID of job at end of pipe chain	*/
  JOBID_U jb ;		/* Job ID union 			*/
  long RetCode ;	/* Return code from Job execution	*/

/* For each command without a piped output */

  for( cno= 0 ; cno < ct->n ; cno++ ) {
    if ( ct->cmd[cno].xopt == EXE_WAIT || ct->cmd[cno].xopt == EXE_FORK ) {


/*

 Note that command is the last in any potential pipe-chain and also note
 that the 'owner' is SB for a background job

*/
      pstat= PCH_LAST ;
      if ( ct->cmd[cno].xopt == EXE_WAIT )
	owner= -1 ;
      else {
	owner= 0 ;
	pstat|= PCH_FORK ;
      }
/*

 For each command in a pipe chain (or just current command if no pipe chain

*/
      for ( i = cno ; !(pstat & PCH_FIRST) && i >= 0 ; i-- ) {

/*

 Determine if command is the first in a pipe-chain, non-chained jobs are
 treated as first & last in a 'chain'

*/
	if ( i == 0 || ct->cmd[i-1].xopt == EXE_WAIT || ct->cmd[i-1].xopt == EXE_FORK )
	  pstat |= PCH_FIRST ;

/* Attempt to load job */

	if ( strlen( ct->cmd[i].name ) > 1 && ct->cmd[i].name[0] == '!' )
	  ct->cmd[i].name++ ;
	if ( ( jid= ct->cmd[i].jobid= jobload(ct->cmd[i].name, ct->cmd[i].arg, owner) ) == -1 ) {
	  qerror2("Unable to load Job", ct->cmd[i].name) ;
	  if ( !( pstat & PCH_LAST ) )
	    jobterm(pejob) ;
	  return ;
	}

/*

 If last (or only) in chain note ID, otherwise pipe to following command

*/
	if ( pstat & PCH_LAST )
	  pejob= jid ;
	else
	if ( jobpipe(jid, ct->cmd[i+1].jobid) ) {
	  qerror2("Cannot connect pipe from Job", ct->cmd[i].name) ;
	  return ;
	}

/*

 If chain is not to become independant assign STDIN and STDOUT to the
 jobs at the ends of the chain and STDERR for each job that doesn't have
 it's error output piped (dup STDOUT for piped error outupt)

*/
	if ( !( pstat & PCH_FORK ) ) {
	  if ( ( pstat & PCH_LAST ) && jobchan(jid, fgetchid(stdout)) ) {
	    qerror2("Cannot assign stdout to Job", ct->cmd[i].name) ;
	    return ;
	  }
	  if ( ct->cmd[i].xopt != EXE_PIPE && jobchan(jid, fgetchid(stderr)) ) {
	    qerror2("Cannot assign stderr to Job", ct->cmd[i].name) ;
	    return ;
	  }
	  if ( ct->cmd[i].xopt == EXE_EPIPE && jobchan(jid, -1L) ) {
	    qerror2("Cannot duplicate stderr for Job", ct->cmd[i].name) ;
	    return ;
	  }
	  if ( ( pstat & PCH_FIRST ) && jobchan(jid, fgetchid(stdin)) ) {
	    qerror2("Cannot assign stdin to Job", ct->cmd[i].name) ;
	    return ;
	  }
	}
	owner= jid ;
	pcno= i ;
	pstat&= ~PCH_LAST ;
      } /* end of <for each command in pipe chain> */

/*

 Start every job in the chain and, for the last job in the chain, report
 the job id for background jobs or the completion code for foreground ones

*/
      for ( i= pcno ; i <= cno ; i++) {
	if ( ct->cmd[i].xopt == EXE_WAIT ) {
	  if ( jobstrt( ct->cmd[i].jobid , 0 , &RetCode ) ) {
	    qerror2("Unable to start Job", ct->cmd[i].name) ;
	    SI(SI_RETCODE, &RetCode) ;
	    return ;
	  }
	}
	else {
	  if ( jobstrt(ct->cmd[i].jobid, 1, &RetCode) ) {
	    qerror2("Unable to start background Job", ct->cmd[i].name) ;
	    SI(SI_RETCODE, &RetCode) ;
	    return ;
	  }
	}
	SI(SI_RETCODE, &RetCode) ;
      }
      if ( rep ) {
	if ( ct->cmd[cno].xopt == EXE_FORK ) {
	  jb.full= ct->cmd[cno].jobid ;
	  printf("Job %d:%d\n", jb.part.entno, jb.part.tag) ;
	}
	else
	if ( RetCode != 0 )
	  printf("Return code: %d\n", RetCode) ;
      }
    }
  } /* end of <for each non-pipe command> */
}
